use crossbeam;
use hashbrown::HashMap;
use itertools::Itertools;
use num_cpus;
use std::io::{self, Read, Write};
use std::iter::FromIterator;
use std::str;

type Frequencies<'a> = HashMap<&'a str, usize>;

fn main() -> io::Result<()> {
    // Read string from stdio
    let mut buffer = String::new();
    io::stdin().read_to_string(&mut buffer)?;
    let buffer_as_bytes = buffer.as_bytes();

    // Figure out optimal chunk size
    let num_threads = num_cpus::get();
    let chunk_size = buffer.len() / num_threads;

    // Figure out positions where we can split string into chunks
    // We only split on whitespaces
    let mut split_positions: Vec<_> = (0..num_threads).map(|i| i * chunk_size).collect();
    split_positions.push(buffer.len());

    let split_positions: Vec<_> = split_positions
        .into_iter()
        .map(|mut pos| {
            while pos < buffer_as_bytes.len() && buffer_as_bytes[pos] == ' ' as u8 {
                pos += 1;
            }
            pos
        })
        .collect();

    // Create references to seperate chunks
    let chunks = (0..num_threads).map(|i| {
        buffer
            .get(split_positions[i]..split_positions[i + 1])
            .unwrap()
    });

    // Tokenize each chunk on a seperate thread
    let frequencies: Result<Vec<_>, _> = crossbeam::scope(|scope| {
        let threads: Vec<_> = chunks
            .map(|chunk| {
                scope.spawn(move |_| {
                    let mut frequencies = Frequencies::new();
                    for word in chunk.split_whitespace() {
                        *frequencies.entry(word).or_insert(0) += 1;
                    }
                    frequencies
                })
            })
            .collect();

        threads.into_iter().map(|t| t.join()).collect()
    })
    .unwrap();

    // Merge results from each thread into a single HashMap
    let frequencies = frequencies.unwrap();

    let frequencies = frequencies
        .into_iter()
        .fold1(merge_frequencies)
        .unwrap_or_default();

    // Sort by value
    let mut frequencies = Vec::from_iter(frequencies);
    frequencies.sort_by(|&(_, a), &(_, b)| b.cmp(&a));

    // Print result on stdout
    let stdout = io::stdout();
    let mut stdout = stdout.lock();
    let mut out = String::new();

    for (word, count) in frequencies {
        let line = format!("{} {}\n", count, word);
        out.push_str(&line);
    }
    stdout.write(out.as_bytes())?;

    Ok(())
}

fn merge_frequencies<'a>(mut a: Frequencies<'a>, b: Frequencies<'a>) -> Frequencies<'a> {
    for (word, count) in b {
        *a.entry(word).or_insert(0) += count
    }
    a
}
